#pragma once

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config_key.h"
#include "driver/gpio.h"
#include "embedded_display.c"
#include "freertos/FreeRTOS.h"
#include "freertos/queue.h"
#include "freertos/task.h"
#include "light.c"

#define GPIO_OUTPUT_IO_0 5
#define GPIO_OUTPUT_IO_1 6
#define GPIO_OUTPUT_PIN_SEL \
  ((1ULL << GPIO_OUTPUT_IO_0) | (1ULL << GPIO_OUTPUT_IO_1))
#define ENCODER_0_CLK_PIN CONFIG_ENCODER_0_CLK_PIN
#define ENCODER_0_DT_PIN CONFIG_ENCODER_0_DT_PIN
#define ENCODER_0_SW_PIN CONFIG_ENCODER_0_SW_PIN
#define ENCODER_0_CLK_PIN_MASK 1ULL << ENCODER_0_CLK_PIN
#define ENCODER_0_DT_PIN_MASK 1ULL << ENCODER_0_DT_PIN
#define ENCODER_0_SW_PIN_MASK 1ULL << ENCODER_0_SW_PIN
#define ENCODER_1_CLK_PIN CONFIG_ENCODER_1_CLK_PIN
#define ENCODER_1_DT_PIN CONFIG_ENCODER_1_DT_PIN
#define ENCODER_1_SW_PIN CONFIG_ENCODER_1_SW_PIN
#define ENCODER_1_CLK_PIN_MASK 1ULL << ENCODER_1_CLK_PIN
#define ENCODER_1_DT_PIN_MASK 1ULL << ENCODER_1_DT_PIN
#define ENCODER_1_SW_PIN_MASK 1ULL << ENCODER_1_SW_PIN
#define ESP_INTR_FLAG_DEFAULT 0

#define NOT_ROTATING -1
#define CLOCKWISE 1
#define COUNTER_CLOCKWISE 0

static const char *UI_INPUT_TAG = "UiInput";

static xQueueHandle ui_input_event = NULL;
static xQueueHandle ui_input_raw_event = NULL;

static int8_t rising_edge_rotation_direction = NOT_ROTATING;
static int8_t falling_edge_rotation_direction = NOT_ROTATING;

uint8_t input_key;

static void IRAM_ATTR gpio_isr_handler(void *arg) {
  input_key = (e_ui_input_raw_key_t)arg;

  switch (input_key) {
    case ui_input_raw_key_encoder_0: {
      s_ui_input_raw_t event = {
          .key = input_key,
          .value = (gpio_get_level(ENCODER_0_SW_PIN) << 2) |
                   (gpio_get_level(ENCODER_0_CLK_PIN) << 1) |
                   (gpio_get_level(ENCODER_0_DT_PIN)),
      };
      xQueueSendFromISR(ui_input_raw_event, &event, NULL);
      break;
    }
    case ui_input_raw_key_encoder_1: {
      s_ui_input_raw_t event = {
          .key = input_key,
          .value = (gpio_get_level(ENCODER_1_SW_PIN) << 2) |
                   (gpio_get_level(ENCODER_1_CLK_PIN) << 1) |
                   (gpio_get_level(ENCODER_1_DT_PIN)),
      };
      xQueueSendFromISR(ui_input_raw_event, &event, NULL);
      break;
    }
    default:
      return;
  }
}

static void ui_input_update_embedded_display(void *arg) {
  s_ui_input_t input;
  char changing_str[12] = "NC";
  for (;;) {
    if (xQueueReceive(ui_input_event, &input, portMAX_DELAY)) {
      switch (input.key) {
        case ui_input_key_display_0_brightness:
          sprintf(changing_str, "Dis0: % 3d", input.value);
          break;
        case ui_input_key_display_1_brightness:
          sprintf(changing_str, "Dis1: % 3d", input.value);
          break;
        case ui_input_key_computer_volume:
          sprintf(changing_str, "CVol: % 3d", input.value);
          break;
        case ui_input_key_display_ambient_lighting_level:
          sprintf(changing_str, "ALLv: % 3d", display_ambient_lighting_level);
          break;
        case ui_input_key_display_ambient_lighting_mode:
          sprintf(changing_str, "ALMd: % 3d", input.value);
          break;
        case ui_input_key_display_0_mode:
          sprintf(changing_str, "Dis0M: % 2d", input.value);
          break;
        case ui_input_key_display_1_mode:
          sprintf(changing_str, "Dis1M: % 2d", input.value);
          break;

        default:
          strcpy(changing_str, "NC");
          break;
      }
      display_fill_rect(0, 6, 128, 8, 0);
      display_print8_str(8, 6, changing_str);
    }
  }
}

static void ui_input_raw_handler(void *arg) {
  s_ui_input_raw_t input;
  uint8_t clk_level;
  uint8_t dt_level;
  uint8_t sw_level;
  int8_t delta = 0;
  char changing_str[12] = "NC";
  for (;;) {
    if (xQueueReceive(ui_input_raw_event, &input, portMAX_DELAY)) {
      sw_level = input.value >> 2 & 1;
      clk_level = input.value >> 1 & 1;
      dt_level = input.value & 1;
      if (clk_level) {
        rising_edge_rotation_direction = dt_level == 0;
        if (falling_edge_rotation_direction == rising_edge_rotation_direction) {
          if (rising_edge_rotation_direction) {
            delta = 1;
          } else {
            delta = -1;
          }
          falling_edge_rotation_direction = NOT_ROTATING;
        } else {
          delta = 0;
          if (falling_edge_rotation_direction != NOT_ROTATING) {
            rising_edge_rotation_direction = NOT_ROTATING;
          }
        }
      } else {
        falling_edge_rotation_direction = dt_level;
        if (rising_edge_rotation_direction == falling_edge_rotation_direction) {
          if (falling_edge_rotation_direction) {
            delta = 1;
          } else {
            delta = -1;
          }
          rising_edge_rotation_direction = NOT_ROTATING;

        } else {
          delta = 0;
          if (rising_edge_rotation_direction != NOT_ROTATING) {
            falling_edge_rotation_direction = NOT_ROTATING;
          }
        }
      }

      if (delta == 0) {
        continue;
      }

      // display_fill_rect(0, 6, 128, 8, 0);

      // switch (input.key) {
      //   case ui_input_raw_key_encoder_0:
      //     sprintf(changing_str, "E0 %d: %u%u%u%u%u", delta,
      //             (input.value >> 4) & 1, (input.value >> 3) & 1,
      //             (input.value >> 2) & 1, (input.value >> 1) & 1,
      //             (input.value) & 1);
      //     break;
      //   case ui_input_raw_key_encoder_1:
      //     sprintf(changing_str, "E1 %d: %u%u%u%u%u", delta,
      //             (input.value >> 4) & 1, (input.value >> 3) & 1,
      //             (input.value >> 2) & 1, (input.value >> 1) & 1,
      //             (input.value) & 1);
      //     break;
      //   default:
      //     strcpy(changing_str, "NC");
      //     break;
      // }
      // display_print8_str(8, 6, changing_str);
      s_ui_input_t event = {.value = delta};
      if (input.key == ui_input_raw_key_encoder_0) {
        if (sw_level) {
          event.key = ui_input_key_computer_volume;
        } else {
          event.key = ui_input_key_display_0_brightness;
        }
      } else if (input.key == ui_input_raw_key_encoder_1) {
        if (sw_level) {
          event.key = ui_input_key_display_ambient_lighting_level;
          led_strip_set_brightness(display_ambient_lighting_level + delta);
        } else {
          event.key = ui_input_key_display_1_brightness;
        }
      }
      xQueueSendFromISR(ui_input_event, &event, NULL);
    }
  }
}

void ui_input_init(void) {
  // zero-initialize the config structure.
  gpio_config_t io_conf = {};

  io_conf.mode = GPIO_MODE_INPUT;
  io_conf.pull_up_en = 1;
  io_conf.pull_down_en = 0;
  // interrupt of rising edge
  io_conf.intr_type = GPIO_INTR_ANYEDGE;
  io_conf.pin_bit_mask = ENCODER_0_CLK_PIN_MASK | ENCODER_1_CLK_PIN_MASK;
  gpio_config(&io_conf);
  io_conf.intr_type = GPIO_INTR_DISABLE;
  io_conf.pin_bit_mask = ENCODER_0_DT_PIN_MASK | ENCODER_1_DT_PIN_MASK;
  gpio_config(&io_conf);
  io_conf.pull_up_en = 1;
  io_conf.pull_down_en = 0;
  io_conf.pin_bit_mask = ENCODER_0_SW_PIN_MASK | ENCODER_1_SW_PIN_MASK;
  gpio_config(&io_conf);

  // start encoder task
  ui_input_event = xQueueCreate(10, sizeof(s_ui_input_t));
  ui_input_raw_event = xQueueCreate(10, sizeof(s_ui_input_raw_t));

  // install gpio isr service
  gpio_install_isr_service(ESP_INTR_FLAG_DEFAULT);
  // hook isr handler for specific gpio pin
  gpio_isr_handler_add(ENCODER_0_CLK_PIN, gpio_isr_handler,
                       (void *)ui_input_raw_key_encoder_0);
  // hook isr handler for specific gpio pin
  gpio_isr_handler_add(ENCODER_1_CLK_PIN, gpio_isr_handler,
                       (void *)ui_input_raw_key_encoder_1);

  xTaskCreate(ui_input_update_embedded_display, "ui_input_event", 2048, NULL,
              10, NULL);
  xTaskCreate(ui_input_raw_handler, "ui_input_event", 2048, NULL, 10, NULL);
}
